from .command import TodoCommand


# Returns the top task from the queue, without tags, a resource, or followons.

class BaseCommand(TodoCommand):

    name = "base"

    @TodoCommand.wrap
    def execute(self):
        return self.queue.base()



from unittest import TestCase
from tempfile import TemporaryDirectory

from busy.handler import main

class TestIntegration(TestCase):

    def test_add_task(self):
        with TemporaryDirectory() as t:
            c = f"--root {t} add --description t"
            main(*c.split())
            with open(f'{t}/tasks.txt') as f:
                l = f.readlines()[-1].strip()
                self.assertEqual(l, 't')



from unittest import TestCase
from tempfile import TemporaryDirectory
from pathlib import Path
from unittest import mock
from io import StringIO
from datetime import date as Date
import unittest

import busy.util.date_util

class TestDate(TestCase):

    @mock.patch('busy.util.date_util.today', lambda : Date(2019,2,11))
    def test_today(self):
        t = busy.util.date_util.relative_date('today')
        self.assertEqual(t, Date(2019, 2, 11))




from unittest import TestCase
import datetime

from busy.model.task import Task
from busy.queue.todo_queue import TodoQueue
from busy.model.task import Plan

# TODO: Clean this up so it doesn't require Task
# TODO: Clean this up so it doesn't depend on defer for other methods

class TestTodoQueue(TestCase):

    def test_top(self):
        q = TodoQueue()
        q.add(Task('a'))
        t = q.top()
        self.assertEqual(str(t),'a')

    def test_list(self):
        s = TodoQueue()
        s.add(Task('a'))
        s.add(Task('b'))
        i = s.list()
        self.assertEqual(len(i), 2)
        self.assertEqual(i[1][0], 2)
        self.assertEqual(str(i[0][1]), 'a')
        self.assertIsInstance(i[1][1], Task)

    def test_defer(self):
        s = TodoQueue()
        s.add(Task('a'))
        d = datetime.date(2018,12,25)
        s.defer(*[1], date=d)
        self.assertEqual(s.plans.count(), 1)
        self.assertEqual(s.plans.top(1).date.year, 2018)

    def test_pop(self):
        s = TodoQueue()
        t1 = Task('a')
        t2 = Task('b')
        s.add(t1)
        s.add(t2)
        s.pop()
        i = s.list()
        self.assertEqual(len(i), 2)
        self.assertEqual(str(i[0][1]), 'b')

    def test_by_number(self):
        s = TodoQueue()
        s.add(Task('a'))
        s.add(Task('b'))
        t = s.top(2)
        self.assertEqual(str(t), 'b')

    def test_create_with_string(self):
        s = TodoQueue()
        s.add('a')
        self.assertEqual(str(s.top()), 'a')
        self.assertIsInstance(s.top(), Task)

    def test_create_with_multiple_strings(self):
        s = TodoQueue()
        s.add('a','b','c')
        self.assertIsInstance(s.top(), Task)
        self.assertEqual(s.count(), 3)
        self.assertEqual(str(s.top(2)), 'b')

    def test_select_multiple(self):
        s = TodoQueue()
        s.add('a','b','c')
        t = s.indices(1,3)
        self.assertEqual(len(t), 2)

    def test_list_plans(self):
        s = TodoQueue()
        s.add('a','b')
        s.defer(0, date=(2018,12,4))
        self.assertEqual(s.plans.count(), 1)

    def test_defer_by_index(self):
        s = TodoQueue()
        s.add('a','b')
        s.defer(1, date=(2018,12,4))
        t = s.top()
        self.assertEqual(s.plans.count(),1)
        self.assertEqual(str(t), 'b')

    def test_defer_multiple(self):
        s = TodoQueue()
        s.add('a','b','c')
        s.defer(1, 3, date=(2018,12,5))
        p = s.plans.top(2)
        self.assertEqual(str(p), 'c')

    def test_list_by_criteria(self):
        s = TodoQueue()
        s.add('a','b','c')
        i = s.list(2,3)
        self.assertEqual(len(i), 2)
        self.assertEqual(str(i[1][1]), 'c')

    def test_drop(self):
        s = TodoQueue()
        s.add('a','b','c')
        s.drop()
        self.assertEqual(str(s.top()), 'b')
        self.assertEqual(str(s.top(3)), 'a')

    def test_drop_by_criteria(self):
        s = TodoQueue()
        s.add('a','b','c','d')
        s.drop('2-3')
        self.assertEqual(str(s.top(1)), 'a')
        self.assertEqual(str(s.top(2)), 'd')
        self.assertEqual(str(s.top(3)), 'b')
        self.assertEqual(str(s.top(4)), 'c')

    def test_activate(self):
        s = TodoQueue()
        s.add('a','b')
        s.defer(1, 2, date=(2018,12,4))
        s.activate(2)
        t = s.top()
        self.assertEqual(str(t), 'b')

    def test_done_no_manager(self):
        s = TodoQueue()
        d = s.done
        self.assertIsNotNone(d)


    # Eliminate blank followons

    def test_finish_output(self):
        s = TodoQueue()
        s.add('a','b')
        f, a, d = s.finish(0, 1, date=(2023,2,12))
        self.assertEqual(a, [])
        self.assertEqual(d, [])


    def test_base(self):
        q = TodoQueue()
        q.add('a at xy #b #c --> repeat tomorrow')
        b = q.base()
        self.assertEqual(b, 'a')

    def test_plans_indices(self):
        q = TodoQueue()
        p = Plan('g',(2023,4,5))
        q.plans.add(p)
        ix = q.plans.indices(1)
        self.assertEqual(ix, [1])


from ..command.command import Command


class CursesCommand(Command):

    name = 'curses'
    ui = 'curses'


    # The execute method might be generic to all interactive UIs.

    def execute(self):
        self._ui.start()
        print(f'Session complete: {self._ui.name}')




from ..command.command import QueueCommand


class ListCommand(QueueCommand):

    name = 'list'
    # key = "l"

    @classmethod
    def set_parser(self, parser):
        super().set_parser(parser)
        self._add_criteria_arg(parser)

    @QueueCommand.wrap
    def execute(self):
        listed = self.queue.list(*self._criteria)
        if len(listed) == 1:
            self.status = f"Listed 1 Item"
        elif listed:
            self.status = f"Listed {str(len(listed))} Items"
        else:
            self.status = "Listed nothing"
        fmtstring = "{0:>6}  " + self.queue.itemclass.listfmt
        return "\n".join([fmtstring.format(i, t) for i, t in listed])





from unittest import TestCase
from unittest.mock import Mock
from tempfile import NamedTemporaryFile

from busy.model.item import Item
from busy.model.item import write_items
from busy.model.item import read_items

class TestItemIO(TestCase):

    def test_writer_tempfile(self):
        i = Mock()
        i.schema = ['d']
        i.d = 'a'
        with NamedTemporaryFile(mode="w+") as tempfile:
            write_items(tempfile, i)
            tempfile.seek(0)
            r = tempfile.readlines()
            self.assertEqual(r, ['a\n'])

    def test_reader_tempfile(self):
        ic = Mock()
        ic.schema = ['d']
        with NamedTemporaryFile(mode="w+") as tempfile:
            tempfile.writelines(['a\n'])
            tempfile.seek(0)
            ix = read_items(tempfile, ic)
            ic.create.assert_called_with({'d':'a'})



from busy.command.command import Command

# Switch to a different queue. Note this only works in interactive UIs; the
# shell UI has no state so it's essentially a no-op.

# TODO: Make SwitchCommand only work in stateful UIs

class SwitchCommand(Command):

    name = "switch"
    key = "w"

    def clean_args(self):
        self._new_queue_name = self._ui.get_string("Switch to queue", "tasks")
        self._new_queue = self._root.get_queue(self._new_queue_name)
        if self.is_omitted('yes'):
            if self._new_queue.count() == 0:
                self.confirm(f"Switch to empty queue {self._new_queue_name}")
            else:
                self._namespace.yes = True


    @Command.wrap
    def execute(self):
        if self._namespace.yes:
            self.queue_name = self._new_queue_name
            self.status = f"Switched to queue {self._new_queue_name}"
        else:
            self.status = "Switch operation canceled"




from unittest import TestCase
from tempfile import TemporaryDirectory
from pathlib import Path
from unittest import mock
from unittest.mock import MagicMock

from busy.queue.todo_queue import TodoQueue
from busy.root.file_system_root import FileSystemRoot
from busy.root.file_system_root import File
from busy.root.file_system_root import InvalidQueueNameError

class TestFileSystemRoot(TestCase):

    def test_root(self):
        with TemporaryDirectory() as d:
            sd = FileSystemRoot(Path(d))
            s = sd.get_queue('tasks')
            self.assertIsInstance(s, TodoQueue)

    def test_add_todo(self):
        with TemporaryDirectory() as td:
            sd1 = FileSystemRoot(Path(td))
            sd1.get_queue('x').add('a')
            sd1.save()
            sd2 = FileSystemRoot(Path(td))
            self.assertEqual(str(sd2.get_queue('x').top()), 'a')

    def test_make_dir_pater(self):
        with TemporaryDirectory() as td:
            r = FileSystemRoot(Path(td))
            r.get_queue('y').add('a')
            r.save()
            r2 = FileSystemRoot(Path(td))
            self.assertEqual(str(r2.get_queue('y').top()), 'a')

    def test_env_var_as_backup(self):
        with TemporaryDirectory() as td:
            with mock.patch.dict('os.environ', {'BUSY_ROOT': td}):
                sd1 = FileSystemRoot()
                sd1.get_queue('p').add('a')
                sd1.save()
                f = Path(td) / 'p.txt'
                self.assertEqual(f.read_text(), 'a\n')

    def test_user_root(self):
        with TemporaryDirectory() as td:
            with mock.patch.dict('os.environ', clear=True):
                with mock.patch('pathlib.Path.home', lambda: Path(td)):
                    sd1 = FileSystemRoot()
                    sd1.get_queue('w').add('a')
                    sd1.save()
                    f = Path(td) / '.busy' / 'w.txt'
                    self.assertEqual(f.read_text(), 'a\n')

    def test_list_queues(self):
        with TemporaryDirectory() as td:
            r = FileSystemRoot(Path(td))
            q1 = r.get_queue('a')
            q1.add('a')
            q2 = r.get_queue('b')
            q2.add('b')
            r.save()
            r2 = FileSystemRoot(Path(td))
            self.assertEqual(len(r2.queue_names), 2)

    def test_queue_not_start_with_dot(self):
        with TemporaryDirectory() as td:
            r = FileSystemRoot(Path(td))
            with self.assertRaises(InvalidQueueNameError):
                q = r.get_queue('.a')
        



from ..command.command import QueueCommand


class DeleteCommand(QueueCommand):

    name = 'delete'
    key = 'd'
    prompt = 'd)elete'

    @classmethod
    def set_parser(self, parser):
        super().set_parser(parser)
        parser.add_argument('--yes', action='store_true')
        self._add_criteria_arg(parser)


    # TODO: Make a convenience method to set the default

    def clean_args(self):
        super().clean_args()
        if self.is_omitted('criteria'):
            self._namespace.criteria = [1]
        if self._indices:
            self._ui.output('\n'.join(self._strings))
            self.confirm(f"Delete {self.count()}")

    # Assume the indices have been already set, before confirmation.

    @QueueCommand.wrap
    def execute(self):
        if self._namespace.yes:
            deleted = self.queue.delete(*self._indices)
            self.status = f"Deleted {self.count(deleted)}"
        else:
            self.status = "Delete command canceled"



from curses import echo, noecho
from busy.util import date_util

from ..command.command import TodoCommand


class DeferCommand(TodoCommand):

    name = 'defer'
    key = 'f'

    @classmethod
    def set_parser(self, parser):
        super().set_parser(parser)
        parser.add_argument('--deferral', '--to', '--for')
        self._add_criteria_arg(parser)


    def clean_args(self):
        def update_deferral():
            deferral = self._ui.get_string("Deferral", "tomorrow")
            self._namespace.deferral = deferral
        super().clean_args()
        if self.is_omitted('criteria'):
            self._namespace.criteria = [1]
        if self._indices:
            self._ui.output('\n'.join(self._strings))
            if self.is_omitted('deferral'):
                self._namespace.deferral = "tomorrow"
            while self.is_omitted('yes'):
                intro = f"Defer {self.count()} to {self._date}"
                self.confirm(intro, update_deferral)

    @property
    def _date(self):
        # if not hasattr(self, '_date_'):
        #     self._date_ = date_util.relative_date(self._namespace.deferral)
        # return self._date_
        return date_util.relative_date(self._namespace.deferral)

    @TodoCommand.wrap
    def execute(self):
        if not self._indices:
            self.status = "Deferred nothing"
        elif not self._namespace.yes:
            self.status = "Defer operation unconfirmed"
        else:
            deferred = self.queue.defer(*self._indices, date=self._date)
            item = deferred[0]
            self.status = f"Deferred {self.count(deferred)} to {item.date}"



import readline
import sys


# Use tab for completion
readline.parse_and_bind('tab: complete')


# Readline defaults to complete with local filenames. Definitely not what we
# want.

def null_complete(text, start):
    return None
readline.set_completer(null_complete)


# I'd like to bind a special key (like ctrl-del) to kill-whole-line, but I can't
# get it to work on my Mac in iTerm. So we're going to use ctrl-A for that.
readline.parse_and_bind('"\C-a": kill-whole-line')


def rlinput(intro:str="", default:str="", options:list=[]):
    """Get input with preset default and/or tab completion of options"""

    # Clean out the options
    options = [o.strip() + " " for o in options]

    # Create the completer using the options
    def complete(text, state):
        results = [x for x in options if x.startswith(text)] + [None]
        return results[state]
    readline.set_completer(complete)

    # Insert the default when we launch
    def start():
        readline.insert_text(default)
    readline.set_startup_hook(start)

    # Actually perform the input
    try:
        value = input(intro)
    finally:
        readline.set_startup_hook()
        readline.set_completer(null_complete)
    return value.strip()



# The Handler is the core control class for a single execution of the busy
# program. It's called when busy is run, and there are some functions to get
# things configured and started. Then one Handler is instantiated. This is the
# only place that imports the UI, Command, and Root classes. So the handler
# instance is passed around, and other classes can access what they need from
# there. Keeping the imports in one place allows the class families to be more
# modular.
#
# A Handler instance provides the following properties. For the most part, these
# properties are how commands access other class families. Remember that the
# Namespace from the Handler is the original Namespace, so use the one passed in
# to the execute() method of the Command for actual usage in an interactive UI.
# Also, the Handler doesn't hold a queue property, because it is only used by
# certain commands. Instead, get your queue from your root yourself!
#
# namespace - The argparse Namespace from the original parsing of the command
# line
#
# root - The Root object for retrieval and storage of items. This is the main
# way that a command gets its root.
#
# ui - The UI instance being implemented. This is the main way a command gets
# its UI.
#
# command - The original Command instance being implemented. Interactive UIs
# might instantiate other Commands.


from argparse import ArgumentParser
import sys
from pathlib import Path
from sys import argv

from busy.command.command import Command
from busy.ui.ui import UserCancelError
from busy.ui.ui import UI
from busy.root.file_system_root import FileSystemRoot
from busy.util.python_version import confirm_python_version

PYTHON_VERSION = (3, 6, 5)

def main(*args):
    if not args:
        args = argv[1:]
    confirm_python_version(PYTHON_VERSION) 
    parser = get_parser()      
    namespace = parser.parse_args(args)
    try:
        handler = Handler(namespace)
        handler.execute()
    except (RuntimeError, ValueError) as error:
        print(f"Error: {str(error)}")

def get_parser():
    parser = ArgumentParser(prog='busy')
    parser.add_argument('--root', action='store', \
        help='file system path to busy data', \
        default=None)
    subparsers = parser.add_subparsers(dest='command')
    for command in Command.family_members('name'):
        key = command.get_member_attr('key')
        aliases = [key] if key else []
        subparser = subparsers.add_parser(command.name, aliases=aliases)
        command.set_parser(subparser)
    return parser


class Handler():

    def __init__(self, namespace):
        self.namespace = namespace
        self.root = FileSystemRoot(namespace.root)
        command_class = Command.family_member('name', namespace.command)
        if not command_class:
            raise RuntimeError(f"Unknown command {namespace.command}")
        self.ui = UI.family_member('name', command_class.ui)(self)
        self.command = command_class(root=self.root, ui=self.ui, \
                                     namespace=self.namespace)

    # Handle the command and output its final status

    def execute(self):
        result = self.command.execute()
        if result:
            print(result)
        if self.command.status:
            print(self.command.status)


    # Some useful methods so other components don't need to import from
    # different class families
    
    def get_commands(self, *attributes):
        return Command.family_members(*attributes)

    def get_command_values(self, *attributes):
        return Command.family_attrs(*attributes)


    # A method for commands within interactive UIs. Use our own root but the UI
    # is passed in. Returns None if attribute value doesn't resolve.
    
    def get_command(self, attribute, value, **kwargs):
        member = Command.family_member(attribute, value)
        if member:
            args = {'root':self.root} | kwargs
            return member(**kwargs)




# The UI to execute one command passed in through the shell. There will be
# limited interactivity, if the user omits an argument on the command line, but
# otherwise this is a run and done situation.
#
# The code for argparse lives at https://github.com/python/cpython/blob/main/Lib/argparse.py


from argparse import ArgumentParser
from io import StringIO

from busy.ui.ui import TerminalUI
from busy.util.readline_util import rlinput
from busy.ui.ui import UserCancelError
from busy.ui.ui import Prompt


class ShellUI(TerminalUI):

    name = "shell"

    def output(self, intro=""):
        """Print some regular output"""
        if intro:
            print(intro)

    def get_string(self, intro, default=""):
        prompt = Prompt(intro=intro)
        return rlinput(intro=self._prompt_string(prompt), default=default)

    def get_option(self, chooser):
        """Get a choice from the user"""
        key = rlinput(intro=self._prompt_string(chooser))
        return chooser.choice_by_key(key)
    
    def _prompt_string(self, prompt):
        result = ""
        if prompt.intro:
            result += prompt.intro + " "
        if prompt.default:
            result += f"[{prompt.default}] "
        for choice in prompt.choices:
            if choice.word==prompt.default:
                continue
            pre, it, post = choice.word.partition(choice.key)
            result += f"{pre}({it}){post} "
        return result + "--> "

    def write_prompt(self, prompt):
        """Output the prompt"""
        print(self._prompt_string(prompt) , end="")





from .command import QueueCommand
import busy


class EditorCommandBase(QueueCommand):

    @classmethod
    def set_parser(self, parser):
        super().set_parser(parser)
        self._add_criteria_arg(parser)


    # Relies on subclasses setting the _default_criterium attribute.

    def clean_args(self):
        super().clean_args()
        if self.is_omitted('criteria'):
            self._namespace.criteria = [self._default_criterium]


    # TODO: Methods to get indices and items as separate lists

    @QueueCommand.wrap
    def execute(self):
        if not self._indices:
            self.status = "Edited nothing"
        else:
            itemclass = self.queue.itemclass
            old = self.queue.items(*self._indices)
            new = self._ui.edit_items(itemclass, *old)
            self.queue.replace(self._indices, new)
            status = f"Edited {self.count(new)}"
            self.status = status


# Special case versions of EditCommand to edit just the top item or the whole
# queue. In the shell you'd just type "edit" or "edit 1-" but this is convenient
# in curses.

class EditCommand(EditorCommandBase):

    name = "edit"
    key = "e"
    _default_criterium = 1

class EditAllCommand(EditorCommandBase):

    name = 'manage'
    key = 'm'
    _default_criterium = "1-"



# Get the top item from a Queue. Returns the entire description, including tags,
# resources, and followons.

from .command import QueueCommand


class TopCommand(QueueCommand):

    name = 'top'

    @classmethod
    def set_parser(self, parser):
        super().set_parser(parser)

    @QueueCommand.wrap
    def execute(self):
        item = self.queue.top()
        if item is None:
            self.status = "No items in queue"
        else:
            return str(item)



import os
from setuptools import setup
from setuptools import find_packages
from pathlib import Path

with open(os.path.join(os.path.dirname(__file__), 'version')) as versionfile:
    version = versionfile.read().strip()

long_description = (Path(__file__).parent / 'README.md').read_text()

setup(name='busy',
      version=version,
      description='Personal time management tool',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='http://gitlab.com/fpotter/tools/busy',
      author='Francis Potter',
      author_email='busy@fpotter.com',
      license='MIT',
      packages=find_packages(),
      entry_points={'console_scripts':['busy=busy.__main__:main']},
      zip_safe=False)



import tkinter as tk  # pragma: no cover


class GetItemWidget(tk.Frame):  # pragma: no cover

    def __init__(self, value=''):
        self.root = tk.Tk()
        self.root.state('zoomed')
        main = tk.Frame(self.root)
        main.pack(fill="both", expand=True)
        frame = tk.LabelFrame(main, text="Item")
        self.entry = tk.Entry(frame, width=100)
        self.entry.insert(0, value)
        self.entry.config(state="readonly", readonlybackground="#D0D0D0")
        self.entry.bind('<Return>', self.enter)
        self.entry.pack()
        self.entry.focus()
        frame.pack(padx=20, pady=20)
        self.root.mainloop()

    def enter(self, event):
        self.root.destroy()



# A terminal-based UI uses Curses, which offers one-key access to most common
# operations.
#
# The terminal has 3 regions, implemented as curses window objects. The regions
# are organized top-to-bottom with a 1-row gap between them. There is also a
# 1-row/1-column gap around the outside of the main terminal window containing
# the regions.
#
# Status region (_statuswin) - 6 lines at the top. Current queue name, current
# task, some blank space, status of the most recent operation, and input prompt
# for the main command.
#
# List region (_listwin) - Takes up the space between the other two regions.
# Shows mostly lists of tasks for confirmation and/or management.
#
# Edit region (_editwin) - 2 lines at the bottom. Used for editing of individual
# task descriptions, such as when adding a new task to the queue.


import curses
from argparse import Namespace
from string import capwords
from time import sleep

from busy.ui.ui import TerminalUI
from busy.ui.ui import UserCancelError
from busy.ui.ui import Prompt
from busy.ui.ui import Chooser


class CursesUI(TerminalUI):  # pragma: nocover

    name = "curses"

    def start(self):
        self._mode = "WORK"
        # Set up the main prompt once for simplicity
        # keynames = self.handler.get_command_values('key', 'name')
        # keynames.add(('q', 'quit'))
        # skeynames = sorted(keynames, key=lambda kn: kn[1])
        # self._main_prompt = Prompt(choices=skeynames)

        # Set up the main command chooser
        chooser = Chooser()
        commands = self.handler.get_commands('key')
        scommands = sorted(commands, key=lambda c:c.name)
        for command in scommands:
            chooser.add_choice(
                keys=[command.key],
                words=[command.name],
                action=command
            )
        chooser.add_choice(keys=['q'], words=['quit'])
        self._command_prompt = chooser

        curses.wrapper(self.term_loop)

    def output(self, intro=""):
        self._descwin.clear()
        self._descwin.addstr(intro)
        self._descwin.refresh()

    def write_prompt(self, prompt):
        """
        Output the prompt with underlines to the window.
        """
        window = self._statuswin
        if prompt.intro:
            window.addstr(prompt.intro + " ", curses.color_pair(3))
        if prompt.default:
            window.addstr(f"[{prompt.default}] ", curses.color_pair(3))
        for choice in prompt.choices:  # TODO: Deal with raw prompt case
            if choice.word==prompt.default:
                continue
            pre, it, post = choice.word.partition(choice.key)
            window.addstr(pre, curses.color_pair(3))
            window.addstr(it, curses.A_UNDERLINE + curses.color_pair(3))
            window.addstr(post + " ", curses.color_pair(3))
        window.addstr("--> ", curses.color_pair(3))


    # TODO: Use a better editing component
    #
    # NOTE: get_string assumes that everything has been cleared and we are in
    # the right place.

    def get_string(self, intro, default=""):
        prompt = Prompt(intro=intro, default=default)
        self._update_status(prompt)
        curses.echo()
        try:
            string = self._statuswin.getstr()
        except KeyboardInterrupt:
            raise UserCancelError
        value = string.decode() or default
        curses.noecho()
        return value

    def get_option(self, chooser):
        """Get a 1-keystroke choice from the user"""
        self._update_status(chooser)
        key = self._get_key()
        return chooser.choice_by_key(key)
    
    def term_loop(self, fullwin):
        self._queue_name = "tasks"
        curses.init_color(curses.COLOR_BLACK, 200, 200, 200)
        curses.init_pair(1, curses.COLOR_CYAN, curses.COLOR_BLACK)
        curses.init_pair(2, curses.COLOR_YELLOW, curses.COLOR_BLACK)
        curses.init_pair(3, curses.COLOR_GREEN, curses.COLOR_BLACK)
        h, w = fullwin.getmaxyx()
        self._statuswin = curses.newwin(5, w-2, 1, 1)
        self._descwin = curses.newwin(h-8, w-2, 7, 1)
        self._editwin = curses.newwin(1, w-2, h-2, 1)
        self._status = "Welcome to Busy!"
        list = ""
        cursor = 0
        while True:
            self.output()
            self._update_status(self._command_prompt)
            try:
                key = self._get_key()
            except UserCancelError:
                break
            if key == "q":
                break
            command = self.handler.get_command('key', key, ui=self, \
                    root=self.handler.root, queue_name=self._queue_name)
            if not command:
                self._status = f"Invalid command {key}"
                continue
            command_name = capwords(command.name)
            self._status = f"{command_name} in progress"
            self._update_status()
            try:
                result = command.execute()
                self._status = command.status or ""
                if hasattr(command, 'queue_name'):
                    self._queue_name = command.queue_name
            except UserCancelError:
                self._status = f"{command_name} command canceled"

    def _update_status(self, prompt=None):
        queue = self.handler.root.get_queue(self._queue_name)
        self._statuswin.clear()
        self._statuswin.move(0, 0)
        self._statuswin.addstr("Queue: ", curses.color_pair(1))
        self._statuswin.addstr(capwords(self._queue_name), curses.A_BOLD)
        self._statuswin.move(1, 0)
        self._statuswin.addstr("Top:   ", curses.color_pair(1))
        self._statuswin.addstr(str(queue.top()) or '', curses.A_BOLD)
        self._statuswin.move(3, 0)
        self._statuswin.addstr(self._status, curses.color_pair(2))
        self._statuswin.move(4, 0)
        if prompt:
            self.write_prompt(prompt)
        cursor = self._statuswin.getyx()
        self._statuswin.refresh()
        self._statuswin.move(*cursor)

    # Convenience method to get one keystroke from the user.
    #     
    def _get_key(self):
        try:
            key = self._statuswin.getkey()
        except (KeyboardInterrupt, curses.error):
            raise UserCancelError
        # self._statuswin.addstr(key)
        return key


    # def listmode(self):
    #     length = len(list.splitlines())
    #     if length:
    #         listpad = curses.newpad(length, curses.COLS)
    #         listpad.addstr(list)
    #         self.listmode = True
    #         listpad.move(cursor, 0)
    #         listpad.addstr(">")
    #         listpad.refresh(0, 0, 5, 0, curses.LINES - 5, curses.COLS)
    #     if self.listmode:
    #         if key == "KEY_UP" and cursor > 0:
    #             cursor = cursor - 1
    #         if key == "KEY_DOWN" and cursor < length:
    #             cursor = cursor + 1
